package SCMxx;

###########################################################################
#
#  gscmxx - a gtk frontend for scmxx
#  Copyright (C) 2002 koppi@weirdoz.net
#  
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#  
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#  
#  You should have received a copy of the GNU General Public License
#  along with this program; if not, write to the Free Software
#  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
# 
#  about scmxx:
#  scmxx is written by Hendrik Sattler <post@hendrik-sattler.de> 
#  find more information at http://www.hendrik-sattler.de/scmxx/
#
#  CHANGELOG
#  2002-11-23:   fixed some other scmxx version specific things
#  2002-11-18:   scmxx >= 0.6.3 uses --out and not --file
#  2002-11-16:   bugfix, splitting of number/name if name contains ','
#  2002-11-04:   pass PREFERENCES as value (hashref) for new SCMxx
#  2002-10-31:   new method: phonedefs returns a hash with detailed
#                information about the phone (bmp, sms, ...)
#  2002-10-19:   '*' and '#' are allowed in phonenumbers
#  2002-10-13:   changed --out to --file (--out will be needed for scmxx > 0.6)
#                new method: delete_addressbook (removes all addresses from phone)
#                save vcard on lowest avail. slot
#  2002-09-13:   let scmxx find a free vcard slot (-F option) and then
#                try to fetch the newly saved vcard (this is slow if you
#                have a lot of vcards)
#                * now the vcard is saved on a free slot, then all free
#                  slots a searched for the specific vcard search 
#                  this is a lot faster!
#  2002-09-03:   added phonebook groups 
#                catch error if RD phonebook is empty
#                added CATEGORIES to vcards
#  2002-08-31:   new function get_model_rev
#                return an error if no vcard could be fetched
#                hash for bitmaps
#  2002-08-30:   added PHONEMODEL and FIRMWAREREVISION
#  2002-08-29:   pod added
#  2002-08-28:   delete a vcard before saving it 
#                (should fix a bug with firmware revision 21 or 24)
#  2002-08-26:   generate_addressbook_from_data should now work with mixed
#                vcards (with and without X-SLOT prop.)
#                property-name FAX fixed => TEL;FAX
#                re-get vcard after saving to show possible changes (cutoff
#                entries) to the user
#  2002-08-25:   SCMxx/Config.pm is only required not a whole package
#  2002-08-22:   vcards do not have empty properties (which make gnomecard 
#                crash)
#  2002-08-21:   new functions create_addressbook_from_data
#                and create_address
#                new poperty X-SLOT in vcard-format
#  2002-08-20:   new function create_phonebook_from_data
#                global var $infotxt (used in various functions)
#                new function setdebug
#                some debug output removed
#
###########################################################################

require 5.005_62;
use strict;
use warnings;

#require Exporter;
#use AutoLoader qw(AUTOLOAD);
#
#our @ISA = qw(Exporter);

require "SCMxx/Config.pm";
$|++;

our $VERSION = "0.4.1";

use vars qw($infotxt);

our %pbookinfos = (
	"FD"	=>	"SIM fix-dialing telephone book (needs SIM PIN2)",
	"SM"	=>	"SIM telephone book",
	"DC"	=>	"Mobile last dialing numbers",
	"MD"	=>	"Mobile last dialing numbers",
	"ON"	=>	"Own telephone numbers in SM+ME",
	"OW"	=>	"Own telephone numbers in SM+ME",
	"LD"	=>	"SIM last dialing numbers",
	"ME"	=>	"Mobile-equipment telephone book",
	"BD"	=>	"Barred (blocked) dialing numbers",
	"SD"	=>	"Service dialing numbers",
	"MC"	=>	"Missed dialing numbers",
	"MS"	=>	"Missed dialing numbers",
	"RC"	=>	"Callback dialling numbers",
	"CD"	=>	"Callback dialling numbers",
	"BL"	=>	"Blacklist dialing numbers",
	"MB"	=>	"Mailbox dialing numbers",
	"CS"	=>	"Common sortable numbers (SM+ME+FD)",
	"RD"	=>	"Red book (VIP) numbers (CS with ! at the end)",
);

our %groups = (
	''	=> 'No Group',
	'!'	=> 'VIP',
	''	=> 'Family',
	''	=> 'Office',
	''	=> 'Leisure',
);

$SCMxx::DEBUG = 0;

#use Data::Dumper;

sub new {
	my $this = shift;
	my %args = (DOIT => '',
				PREFERENCES => $SCMxx::Config,
				@_);
	my $class = ref($this) || $this;
	my $self = {};
	bless $self, $class;
	# needed for my gtk-interface (gscmxx)
	$self->{DOIT} = $args{DOIT};
	$self->initialize(PREFERENCES => $args{PREFERENCES});
	return $self;
}  


##################################################
# set initial params
##################################################
sub initialize {
	my $self = shift;
	my %args = (PREFERENCES => '', 
				@_);
	print "Init SCMxx.pm" if $SCMxx::DEBUG;
	foreach (keys %{$SCMxx::Config}) {
		if ($args{PREFERENCES}->{lc($_)}) {
			$self->{uc($_)} = $args{PREFERENCES}->{lc($_)};
		} else {
			$self->{uc($_)} = $SCMxx::Config->{uc($_)};
		}
	}
	# version of scmxx
	my $result = $self->docmd("--version");
	chomp $result;
	if ($result =~ /ERROR/) {
		$self->{VERSION} = 0;
	} else {
		$self->{VERSION} = $result;
	}
}

sub set_preferences {
	my $self = shift;
	my $prefs = shift;
	foreach (keys %$prefs) {
		$self->{uc($_)} = $prefs->{$_};
	}
	$infotxt = "";
	# version of scmxx
	my $result = $self->docmd("--version");
	chomp $result;
	if ($result =~ /ERROR/) {
		$self->{VERSION} = 0;
	} else {
		$self->{VERSION} = $result;
	}
}

sub get_model_rev {
	my $self = shift;
	my $info = $self->information;
	foreach (split /\n/, $info) {
		chomp;
		if (/^Revision/i) {
			my (undef, $rev) = split (/\s+/, $_, 2);
			$self->{FIRMWAREREVISION} = $rev;
		}
		if (/^Model/i) {
			my (undef, $mod) = split (/\s+/, $_, 2);
			$self->{PHONEMODEL} = $mod;
		}
	}
	$self->{FIRMWAREREVISION} = 0;
	$self->{PHONEMODEL} = "Unknown";
}

##################################################
# INFORMATION ABOUT THE PHONE 
##################################################
sub information {
	my $self = shift;
	return $infotxt if ($infotxt);
	$infotxt = $self->docmd("-i --file - 2>&1"); 
	$infotxt =~ s/Phonebooks:/\nPhonebooks:/;
	$infotxt =~ s/SMS storages:[\t\s\n\r]+(.*)/\nSMS storages:\t$1/;
	return $infotxt;
}

##################################################
# DETAILED INFORMATION ABOUT THE PHONE 
##################################################
sub phonedefs {
	my $self = shift;
	print "Getting Phone Defs ...\n" if $SCMxx::DEBUG;
	my $info = $self->information;
	my @mobileinfo;
	if ($info =~ /^ERROR /) {
		return ($info, {});
	} else {
		@mobileinfo = split(/\n/, $info);
	}
	my %phonedefs = ();
	foreach (@mobileinfo) {
		my ($k, $v) = split (/:\s+/, $_, 2);
		next unless ($k && $v);
		if ($k eq "Avail. Slots" || $k eq "Readable Slots") {
			my @avail = split (/,\s+/, $v);
			foreach (@avail) {
				my ($type, $num) = split (/: /, $_);
				$phonedefs{slots}->{$type} = $num;
			}
		} elsif ($k eq "Bitmap (bmp)") {
			$phonedefs{slots}->{bmp} = $v;
		} elsif ($k eq "Midi (mid)") {
			$phonedefs{slots}->{mid} = $v;
		} elsif ($k eq "vCard (vcf)") {
			$phonedefs{slots}->{vcf} = $v;
		} elsif ($k eq "vCal (vcs)") {
			$phonedefs{slots}->{vcs} = $v;
		} else {
			$phonedefs{$k} = $v;
		}
	}
	return (0, \%phonedefs);
}

##################################################
# RESET THE PHONE
##################################################
sub reset {
	my $self = shift;
	my $result = $self->docmd("--reset --file - 2>&1");
	$infotxt = "";
	return $result;
}


##################################################
# RETREIVE SMS FROM THE PHONE 
##################################################
sub get_sms {
	my $self = shift;
	my $slot = shift || "all";
	my $format = shift || "ARRAY";

	my $smstxt = $self->docmd("--get --sms=$slot --file - 2>&1"); # /dev/null");
	my @tmp = split(/\n/, $smstxt);
	$smstxt = "";
	foreach (@tmp) {
		next if (/^Looking for/ || /^Receiving /);
		$smstxt .= $_ . "\n";
	}
	if (uc($format) eq "TEXT") {
		return $smstxt;
	} else {
		my %sms; my @sms;
		if ($smstxt =~ /There are no SMS of this type/) {

		} else {
			my @smsdata = split (/\n/, $smstxt);
			my $slot;
			my $tmpslot = 0;
			my $tmp = "";
			foreach my $l (@smsdata) {
				if ($l =~ /^Slot: (\d+)$/) {
					$tmpslot = $1;
					if ($tmp && $slot) {
						if (uc($format) eq "HASH") {
							$sms{$slot} = &read_one_sms($tmp);
						} else {
							push @sms, &read_one_sms($tmp);
						}
					}
					$tmp = "";
					$slot = $tmpslot;
				} # else {
					$tmp .= $l . "\n";
				#}
			}
			if ($format eq "HASH") {
				$sms{$slot} = &read_one_sms($tmp) if $tmp && $slot;
			} else {
				push @sms, &read_one_sms($tmp) if $tmp && $slot;
			}
		}
		if ($format eq "HASH") {
			return (wantarray) ? %sms : \%sms;
		} else {
			return (wantarray) ? @sms : \@sms;
		}
	}
}

sub read_one_sms {
	my $tmp = shift;

	return unless $tmp;

	my %sms;
	$tmp =~ s/\x0D//g;
	my @tmp = split (/\n/, $tmp);
	my $ismessage = 0;
	foreach (@tmp) {
		if ($_ eq "Message:") {
			$ismessage = 1;
			next;
		}

		if ($ismessage) {
			$sms{Message} .= $_ . "\n";
		} else {
			my ($k, $v) = split(/: /, $_, 2);
			$sms{$k} = $v if ($k && $v);
		}
	}
	if (exists $sms{To}) {
		$sms{Type} = "out";
	} else {
		$sms{Type} = "in";
	}
	return \%sms;
}

##################################################
# DELETE SMS FROM THE PHONE 
##################################################
sub delete_sms {
	my $self = shift;
	my $slot = shift;
	return "No slot selected!" unless defined $slot;

	my $sms = $self->get_sms($slot, "TEXT");
	if ($sms) {
		my $result = $self->docmd("--remove --sms=$slot --file - 2>&1");
		return $result;
	} else {
		return "SMS not found!";
	}
}

##################################################
# SEND SMS (optional flash-sms) 
##################################################
sub send_sms {
	my $self = shift;
	my $number = shift;
	my $text = shift || "";
	my $flash = shift;
	my $save = shift;

	# format text
	$text =~ s/\n/ /g;
	$text =~ s/\s+/ /g;
	$text =~ s/"/\\"/g; 

	my $error;
    if (!$number) {
        $error .= "Please insert Number to send to!\n";
    } else {
		my $tmperr;
		($number, $tmperr) = __control_number($number);
		$error .= $tmperr . "\n" if ($tmperr);
    }
    if (!$text) {
        $error .= "No text given, will not send emtpy SMS!\n";
    }

	if ($error) {
		return (0, "$error");
	}

	my $options = qq|--send --sms --number $number --text "$text" |;
	$options .= qq| --direct | unless $save;
	$options .= ($flash) ? " --flash" : '';
	$options .= " 2>&1";
	my $result = $self->docmd($options);
	if ($result =~ /^SMS text is too long/) {
		return (0, $result);
	}
	return (1, $result);	
}




##################################################
# RETREIVE PHONEBOOK FROM PHONE
##################################################

sub get_phonebook {
	my $self = shift;
	my $pbook = shift || "SM"; # use SIM Card Phonebook as default
	my $format = shift || "ARRAY";	
	my $timeout = shift || 30;
	$pbook = uc($pbook);
	my $tmptimeout = $self->{TIMEOUT}; 
	$self->{TIMEOUT} = $timeout;
	my $result = $self->docmd("--get --pbook=$pbook --file - 2>&1");
#	my $result = `cat /tmp/pbook/`;
	$self->{TIMEOUT} = $tmptimeout;

	return $result if $format eq "TEXT";

	if ($result =~ /^ERROR/) {
		# VIP numbers
		if ($pbook eq "RD" && $result =~ /Error on selecting phonebook/) {
			return $self->create_phonebook_from_data("",$format);
		} else {
			my @tmp = grep !/^\d+,/, split (/\n/, $result);
			$result = join ("\n", @tmp);
			return $result;
		}
	}
	return $self->create_phonebook_from_data($result,$format);
}

sub create_phonebook_from_data {
	my $self = shift;
	my $data = shift;
	my $format = shift;

	my %pbook;
	my @pbook;
	my $count;
	my @lines = split (/\n/, $data);
	foreach (@lines) {
        if ($_ =~ /^\d+,/) {
            my ($i, $number, $name) = split(/"?,"/, $_);
            $name =~ s/^"//;
            $name =~ s/"$//;
            $number =~ s/^"//;
            $number =~ s/"$//;
			my $group = '';
			foreach my $gr(keys %groups) {
				next unless $gr;
				if ($name =~ /$gr$/) {
					$name =~ s/$gr$//;
					$group = $gr;
				}
			}
			if ($format eq "HASH") {
				$pbook{$i} = { NAME => $name, NUMBER => $number, GROUP => $group };
			} else {
				push @pbook, { NAME => $name, NUMBER => $number, GROUP => $group };
			}
			$count++;
        }
    }
	if ($format eq "HASH") {
		$pbook{__count} = $count;
		return (wantarray) ? %pbook : \%pbook;
	} else {
		return (wantarray) ? @pbook : \@pbook;
	}
}
##################################################
# get available phonbooks (and infos about them)
##################################################

sub available_phonebooks {
	my $self = shift;
	$infotxt ||= $self->information();
	my @info = split (/\n/, $infotxt);
	my $pbooks = (grep /^Phonebooks:/, @info)[0] || "";
	$pbooks =~ s/^Phonebooks:\s+//;

	my %pbooks;
	foreach (split (/, /, $pbooks)) {
		$pbooks{$_} = $pbookinfos{$_} || "Unknown Phonebook";		
	}
	delete $pbooks{DC} if ($pbooks{MD});
	delete $pbooks{ON} if ($pbooks{OW});
	delete $pbooks{MC} if ($pbooks{MS});
	delete $pbooks{RC} if ($pbooks{CD});

	return (wantarray) ? %pbooks : \%pbooks;
}

##################################################
# generate a phonebook in scmxx-readable format
##################################################

sub generate_phonebook {
	my $self = shift;
	my $pbook = shift;

	return unless $pbook;
	if (!ref $pbook) {
		my @pbook = split (/\n/, $pbook);
		return $self->generate_phonebook(\@pbook);
	} elsif (ref($pbook) eq "ARRAY") {
		my $tmp;
		my $i = 1;
		foreach my $p (@{$pbook}) {
			if (!ref($p)) {
				if ($p !~ /^\d+,"[+\s\d]*","[^"]*"$/) {
					return 0, "Phonebook Format is not OK: $p\n";
				} else {
					$p =~ s/^(\d+),/$i,/;
					$tmp .= $p . "\n";
					$i++;
				}
			} elsif (ref($p) eq "HASH") {
				if (defined $p->{NUMBER} || defined $p->{NAME}) {
					$p->{NUMBER} =~ s/"/\"/g;
					$p->{NAME} =~ s/"/\"/g;
					my $name = $p->{NAME} || "";
					$name .= $p->{GROUP} if ($p->{GROUP}); #  && $p->{GROUP} eq "VIP");
					$tmp .= qq|$i,"$p->{NUMBER}","$name"\n|;
					$i++;
				}
			} elsif (ref($p) eq "ARRAY") {
#=comment TODO
#				my @ptmp = @{$p};
#				if ($#ptmp > 2) {
#					die "Phonebook Format is not OK: to much columns\n" . join (",", @ptmp) . "\n",
#				}
#				if ($#ptmp == 2) {
#					
#				}
#=cut
			}
		}
		return 1, $tmp;
	} elsif (ref($pbook) eq "HASH") {
		my $tmp;
		my $i = 1;
		foreach (keys %{$pbook}) {
			next if /^__/;
			my $p = $pbook->{$_};
			if (!ref($p)) {
				return (0, "Phonebook Format is not OK: not a hash ref");
			}
			if (defined $p->{NUMBER} || defined $p->{NAME}) {
				$p->{NUMBER} =~ s/"/\"/g;
				$p->{NAME} =~ s/"/\"/g;
				my $name = $p->{NAME} || "";
				$name .= $p->{GROUP} if ($p->{GROUP});
				$tmp .= qq|$i,"$p->{NUMBER}","$name"\n|;
				$i++;
			}
		}
		if ($pbook->{'__count'}) {
			foreach ($i..$pbook->{'__count'}) {
				$tmp .= qq|$_,"",""\n|;
			}
		}
		return 1, $tmp;		
	} else {
		return (0, "Phonebook Format is not OK: data MUST be a 'SCALAR' or 'ARRAY'- or 'HASH'-REFERENCE'");
	}
}

##################################################
# SAVE PHONEBOOK
##################################################

sub save_phonebook {
	my $self = shift;
	my $pbook = shift;
	my $data = shift;
	my $pin2 = shift;

	return unless $pbook && $data;
	my ($status, $pdata) = $self->generate_phonebook($data);

	if ($status == 0) {
		return $pdata;
	}

	my $tmp = __find_tmp_dir();

	print "Using tmp dir: $tmp\n" if $SCMxx::DEBUG > 1;

	my $tmpfile = $tmp . "/pbook.tmp";
	open (PBOOK, ">$tmpfile") || die "Can't open $tmpfile $!";
    print PBOOK $pdata;
    close (PBOOK);
	my $timeout = $self->{TIMEOUT}; 
	$self->{TIMEOUT} = $self->{PBTIMEOUT};
	my $options = "-v --send --pbook=$pbook $tmpfile ";
	$options .= "--pin $pin2 " if $pin2;
	$options .= "2>&1"; 
	my $result = $self->docmd($options);
	$self->{TIMEOUT} = $timeout;
#	unlink $tmpfile;	
	my @tmp = split /\n/, $result;
	my $tmpresult;
	foreach (@tmp) {
		next if (/^Sending/ || /^Current/ || /^Received/ || 
				 /^Command sent/ || /^\d+ Current / || /^$/ ||
				 /^OK, a modem/ || /^Detected/ || /^Preparing/ ||
				 /^Accessing /);
		$tmpresult .= $_ . "\n";
	}
	return $tmpresult;
}




##################################################
# RETREIVE ADDRESSBOOK FROM PHONE
##################################################

sub get_addressbook {
	my $self = shift;
	my $timeout = shift || 60;
	my $tmptimeout = $self->{TIMEOUT}; 
	$self->{TIMEOUT} = $timeout;
	my $result = $self->docmd("--get --vcard=all --file - 2>&1");
#	my $result = ` cat /tmp/adrbook.txt` ;
	$self->{TIMEOUT} = $tmptimeout;



	$result =~ s/(?<!END:VCARD\n)File transfer complete./\nEND:VCARD\n/g;
	$result =~ s/Receiving packet \d+ of \d+\.\.\.\n//g;
	$result =~ s/File transfer complete.//g;
#+	my %addressbook;
	if ($result =~ /could not be retrieved/ || $result =~ /Error executing/i) {
		my @tmp = split (/\n/, $result);
		$result = "";
		foreach (@tmp) {
			next if (/^Slot / || /^BEGIN:/ || /^VERSION/ || /^REV:/ || /^FN:/ || 
					 /^N:/ || /^TEL;/ || /^END/ || /^ADR:/ || /^ORG:/ || /^EMAIL;/ ||
					 /^URL:/ || /^FAX/ || /^File transfer complete/ || /^Empty slot:/);
			$result .= $_ . "\n";
		}
		return (0,$result);
#		$addressbook{__error} = "Error:\n$result\n";
	}

#	print $result;

	my $addressbook = $self->create_addressbook_from_data($result);

#	print Dumper($addr);

#	%addressbook = %{$addr};

	$addressbook->{"__fetched"}->{FN} = 1;
	return (1, $addressbook);
#	return (wantarray) ? %addressbook : \%addressbook;
}	


sub delete_addressbook {
    my $self = shift;
    my $timeout = shift || 60;
    my $tmptimeout = $self->{TIMEOUT};
    $self->{TIMEOUT} = $timeout;
    my $result = $self->docmd("--remove --vcard=all --file - 2>&1");
    $self->{TIMEOUT} = $tmptimeout;
	return $result;

}
sub create_addressbook_from_data {
	my $self = shift;
	my $data = shift;

	# fix not fully retrieved vcards
	$data =~ s/[\r\n]+/\n/g;
	$data =~ s/E\w*BEGIN:VCARD/END:VCARD\nBEGIN:VCARD/g;
	$data =~ s/(?<!END:VCARD\n)BEGIN:VCARD/END:VCARD\nBEGIN:VCARD/g;
	$data =~ s/END:VCARD\n(Slot \d+\.\.\.)\nEND:VCARD\nBEGIN:VCARD/END:VCARD\n$1\nBEGIN:VCARD/g;

	my @result = split (/[\r\n]+/, $data);
	my $slot;
	my $tmp;
	my %addressbook;
	my $i = 0;
	foreach my $line (@result) {
		chomp $line;
		my $tmpslot;
		if ($line =~ /^Slot (\d+)/) {
			$slot = $1;
		}
		elsif ($line =~ /^X-SLOT:(\d+)/) {
			$slot = $1;
		}
		elsif ($line =~ /Empty slot/) {
			undef $slot;
		}
		elsif ($line =~ /^END/) {
			print "\n-----------\nSLOT: $slot\n$tmp\n---------------\n" if $SCMxx::DEBUG > 1;
			if ($tmp) {
				$tmp .= $line;
				$tmp =~ s/\x0D//g;
				my $myslot;
				if (defined $slot) {
					$myslot = $slot;
				} else {
					$myslot = "NEW_$i";
					$i++;
				}
				undef $slot;
				print "\n\n$tmp\n\n" if $SCMxx::DEBUG > 1;
				my $adr = $self->create_address($tmp);	
				$addressbook{$myslot} = $adr if $adr; 
			}
			$tmp = "";
		} else {
			$tmp .= $line . "\n";
		}
		if ($tmpslot && $slot && $tmpslot != $slot) {
			$slot = $tmpslot;
		}
	}
	my $myslot;
	if (defined $slot) {
		$myslot = $slot;
	} else {
		$myslot = "NEW_$i";
		$i++;
	}
	print "\n\n$tmp\n\n" if $SCMxx::DEBUG > 1;
	my $adr = $self->create_address($tmp);
	$addressbook{$myslot} = $adr if $adr;

	my @xslots = reverse sort grep !/^NEW/, keys %addressbook;
	my $maxslot = $xslots[0];
	foreach (sort grep /^NEW/, keys %addressbook) {
		print "$_ ... " if $SCMxx::DEBUG > 1;
		$addressbook{++$maxslot} = delete $addressbook{$_};
		print "$maxslot\n" if $SCMxx::DEBUG > 1;
	}
	return \%addressbook;
}

sub create_address {
	my $self = shift;
	my $data = shift;

	return 0 unless $data;
	my %address;
	foreach my $l(split(/\n/, $data)) {
		chomp ($l);
		next if ($l =~ /Empty slot/);
		next if ($l =~ /File transfer complete/);
		my ($prop, $val) = split (/:/, $l, 2);
		next unless $prop;
		next if ($prop eq "BEGIN" || $prop eq "END");
		if ($prop && $val) {
			# name
			if ($prop eq "N") {
				($address{N}, $address{SN}) = split(";", $val);
			# address
			} elsif ($prop eq "ADR") {
				(undef, undef, $address{STREET}, $address{CITY}, 
				undef, $address{ZIP}, $address{COUNTRY}) =  split (/;/, $val || "");
			# full name
			} elsif ($prop eq "FN") {
				$val =~ s/^\s*,\s+//;
				$address{$prop} = $val;
			} else {
				$address{$prop} = $val;
			}
		}
	}
#	use Data::Dumper;
#	print Dumper(\%address);
	if (keys %address) {
		return \%address;
	} else {
		return 0;
	}
}

sub delete_address {
	my $self = shift;
	return $self->delete_addressbook_entry(@_);
}

sub delete_addressbook_entry {
	my $self = shift;
	my $slot = shift;

	my $result = $self->docmd("--remove --vcard=$slot --file - 2>&1");
	print $result if $SCMxx::DEBUG > 1;
	return $result;
}

sub save_address {
	my $self = shift;
	my $slot = shift;
	my $address = shift;
	my $slots = shift;

	my %posslots = map { $_ => 1 } (0..500);
	delete $posslots{$_} foreach @{$slots};
	$posslots{$slot}++ if $slot;
	my $newslot = (sort { $a <=> $b } keys %posslots)[0];
	print "New Slot: $newslot\n" if $SCMxx::DEBUG > 1;

	my $vcard;
	($address, $vcard) = $self->generate_vcard($address, $slot);

	if ($vcard !~ /^BEGIN:VCARD/) {
		print "Returning error $vcard " if $SCMxx::DEBUG > 1;
		return ($vcard, 0, $address);
	}

	(my $fn = $address->{FN}) =~ s/\W//g;
	$fn .= "_" . $address->{REV} . ".vcard";
	print "$fn\n" if $SCMxx::DEBUG > 1;

	my $tmp = __find_tmp_dir();
	my $tmpfile = "$tmp/$fn";
	open (VCARD, ">$tmpfile") || die "Can't open $tmpfile $!";
	print VCARD $vcard;
	close (VCARD);

	my $timeout = $self->{TIMEOUT};
	$self->{TIMEOUT} = $self->{ABTIMEOUT} || 60;

	# get model and firmware revision if not detected yet
	unless ($self->{PHONEMODEL} && $self->{FIRMWAREREVISION}) {
		$self->get_model_rev();	
	}
	# first delete the vcard 
	# (on s45 wit REV >= 21 the vcard was be saved twice)
#	if ($self->{PHONEMODEL} eq "S45" && $self->{FIRMWAREREVISION} >= 21) {
	if ($slot) {
		my $rresult = $self->delete_address($slot);
		print "DELETE: $rresult\n" if $SCMxx::DEBUG;
		print "Sleeping for 2 second ... \n" if $SCMxx::DEBUG;
		sleep 2;
		$posslots{$slot} = 1;  # on this slot a vcard can be saved again
		if ($rresult =~ /ERROR/) {
			$self->{TIMEOUT} = $timeout;
			return ($rresult, 0, $address)
		}
	}


	# actually save the vcard
	# my $result = $self->docmd("--send -F $tmpfile 2>&1");
	my $result = $self->docmd("--send --vcard=$newslot $tmpfile 2>&1");
#	unlink($tmpfile);
	print "SEND: $result\n\n" if $SCMxx::DEBUG;

#	$result =~ /Detected empty slot (\d+)/;
#	my $newslot = $1 || "";
#	print "NEW SLOT: $newslot\n\n" if $SCMxx::DEBUG;
	#if ($result =~ /ERROR/ || !$newslot) {
	if ($result =~ /ERROR/) {
		$self->{TIMEOUT} = $timeout;
		return ($result, $newslot, $address);
	}


	my $addr;
	my $fullname = $address->{FN};

	# fetch address from phonebook to verify entries
#	if ($newslot) {
		my $xresult = $self->docmd("--get --vcard=$newslot --file - 2>&1");
		$xresult =~ s/(?<!END:VCARD\n)File transfer complete./\nEND:VCARD\n/g;
		$xresult =~ s/Receiving packet \d+ of \d+\.\.\.\n//g;
		$xresult =~ s/File transfer complete.//g;
		print "GET:\n$xresult\n\n" if $SCMxx::DEBUG > 1;
		delete $posslots{$newslot};
		if ($xresult =~ /ERROR/ || $xresult =~ /nothing to get/) {
	#		$self->{TIMEOUT} = $timeout;
	#		return ($xresult, 0, $address);
		} else {
			$addr = $self->create_address($xresult);
			if ($addr->{FN} eq $fullname) {
				$self->{TIMEOUT} = $timeout;
				return ("VCard saved", $newslot, $addr);
			}
		}
#	}


	# try to find the address 
	foreach my $sl (sort { $a <=> $b } keys %posslots) {
		print "Trying SLOT: $sl\n" if $SCMxx::DEBUG;
		my $xresult = $self->docmd("--get --vcard=$sl --file - 2>&1");
		$xresult =~ s/(?<!END:VCARD\n)File transfer complete./\nEND:VCARD\n/g;
		$xresult =~ s/Receiving packet \d+ of \d+\.\.\.\n//g;
		$xresult =~ s/File transfer complete.//g;
		print "GET:\n$xresult\n\n" if $SCMxx::DEBUG > 1;
		if ($xresult =~ /ERROR/ || $xresult =~ /nothing to get/) {
	#		$self->{TIMEOUT} = $timeout;
	#		return ($xresult, 0, $address);
		} else {
	    	$addr = $self->create_address($xresult);
			if ($addr->{FN} eq $fullname) {
				$self->{TIMEOUT} = $timeout;
				return ("VCard saved", $sl, $addr);
			}
		}
	}

	# never should come here ....
	return ("ERROR\nVCard could not be fetched from the phone ....", 0, $address);
}

sub generate_vcard {
	my $self = shift;
	my $address = shift;
	my $slot = shift; 

	$slot = "" unless defined $slot;

	my @keys = ( "N", "SN", "ORG", "EMAIL;INTERNET", "TEL;HOME", "TEL;WORK", "TEL;CELL", "TEL;FAX", "STREET", "ZIP", "CITY", "COUNTRY", "URL");
	foreach (@keys) {
		$address->{$_} ||= "";
	}

	my $error;
	foreach ("TEL;HOME", "TEL;WORK", "TEL;CELL", "TEL;FAX") {
		my $tmperr;
		($address->{$_}, $tmperr) = __control_number($address->{$_}) if ($address->{$_});
		$error .= $tmperr . "\n" if $tmperr;
	}

	if ($error) {
		return ($address, $error);
	}
	# the phone set FN to Lastname, Firstname :-(
#	unless ($address->{FN}) {
		$address->{FN} = "";
		$address->{FN} = $address->{N} if ($address->{N});
		$address->{FN} .= ", " if ($address->{N} && $address->{SN});
		$address->{FN} .= $address->{SN} if ($address->{SN});
		$address->{FN} ||= "Unknown";
#	}
	print "FN: $address->{FN}\n" if $SCMxx::DEBUG > 1;

	my @lt = localtime(time);
	$lt[5]+=1900;
	$lt[4]++;
	my $rev = sprintf("%4d%02d%02dT%02d%02d%02d", @lt[5,4,3,2,1,0]);
	$address->{REV} = $rev;
	# create vcard
	my $vcard = qq|BEGIN:VCARD
VERSION:1.0
REV:$rev
FN:$address->{FN}\n|;
	my $name = $address->{N} . ";" . $address->{SN};
#	$name =~ s/^\s*;\s*//; 
#	$name =~ s/\s*;\s*$//; 

	$vcard .= "N:$name\n" if $name;

	if ($address->{STREET} || $address->{CITY} || $address->{ZIP} || $address->{COUNTRY}) {
		$vcard .= "ADR:;;$address->{STREET};$address->{CITY};;$address->{ZIP};$address->{COUNTRY}\n";
	}
	$vcard .= "CATEGORIES:$address->{CATEGORIES}\n" if $address->{CATEGORIES};
	$vcard .= "ORG:$address->{ORG}\n" if $address->{ORG};
	$vcard .= "TEL;HOME:$address->{'TEL;HOME'}\n" if $address->{'TEL;HOME'};
	$vcard .= "TEL;WORK:$address->{'TEL;WORK'}\n" if $address->{'TEL;WORK'};
	$vcard .= "TEL;CELL:$address->{'TEL;CELL'}\n" if $address->{'TEL;CELL'};
	$vcard .= "TEL;FAX:$address->{'TEL;FAX'}\n" if $address->{'TEL;FAX'};
	$vcard .= "EMAIL;INTERNET:$address->{'EMAIL;INTERNET'}\n" if $address->{'EMAIL;INTERNET'};
	$vcard .= "URL:$address->{URL}\n" if $address->{URL};
	$vcard .= "X-SLOT:$slot\n";
	$vcard .= "END:VCARD\n";

	print "-------------\n$vcard\n---------------------\n" if $SCMxx::DEBUG > 1;

	return ($address, $vcard);
}



##################################################
# GET BITMAPS FROM THE PHONE
##################################################
sub get_bitmaps {
	my $self = shift;
	my $slot = shift || "all";
	my $bitmapdir = shift || ".";
	my $format = shift || "";

    # create bitmap dir if not exists
    if (!-e "$bitmapdir") {
        mkdir ($bitmapdir, 0750);
    } else {
        # remove old bitmap files
		foreach (0..4) {
			system("rm -f $bitmapdir/$_.bmp");
		}
    }

    my $result = $self->docmd("--get --bitmap=$slot --file $bitmapdir/ 2>&1");
	my @tmp = split (/\n/, $result);
#	my @bitmaps = grep (/created/, @tmp);
#	foreach (@bitmaps) {
	my %bitmaps;
	foreach (grep (/created/, @tmp)) {
		chomp;
		s/^.*\///;
		s/\s+created.//;
		my $bitmap = $_;
		$bitmap =~ /(\d+)\.bmp/;
		$bitmaps{$1} = $bitmap;
	}
	# convert to xpm
	if ($format eq "xpm") {
#		print "Converting to xpm .... \n";
		foreach my $i (keys %bitmaps) {
			my $bmp = $bitmaps{$i};
			my $xpm = $bmp;
			$xpm =~ s/bmp/xpm/;
			my $cmd = "bmptoppm $bitmapdir/$bmp 2>/dev/null | ppmtoxpm > $bitmapdir/$xpm 2>/dev/null";
			system($cmd);
			$bitmaps{$i} = $xpm;
#			print "     $bitmapdir/$xpm\n";
		}
	}
	return (wantarray) ? %bitmaps : \%bitmaps;
}

##################################################
# GET MIDI FILES FROM THE PHONE
##################################################

sub get_sounds {
	my $self = shift;
	my $slot = shift || "all";
	my $dir = shift || ".";
	my $format = shift;

    # create bitmap dir if not exists
    if (!-e "$dir") {
        mkdir ($dir, 0750);
    } else {
        # remove old bitmap files
		foreach (0..4) {
	        system("rm -f $dir/$_.mid");
		}
    }

    my $result = $self->docmd("--get --midi=$slot --file $dir/ 2>&1");
	my @tmp = split (/\n/, $result);
	my @sounds = grep (/created/, @tmp);
	foreach (@sounds) {
		chomp;
		s/^.*\///;
		s/mid\s+created.*$/mid/;
	}
	return @sounds;
}






##################################################
# actually send a command to the phone (scmxx)
##################################################
sub docmd {
	my $self = shift;
	my $options = shift;
	## scmxx >= 0.6.3 uses --out and not --file
	if ($self->{VERSION}) {
		my @ver = split /\./, $self->{VERSION};
		$ver[1] ||= 0;
		$ver[2] ||= 0;
		$ver[3] ||= 0;
		my $ver = int($ver[0] . sprintf("%03d", $ver[1]) . sprintf("%03d", $ver[2])  . sprintf("%03d", $ver[3])); 
		if ($ver > 6002000) {
			$options =~ s/--file/--out/g;
		}
	}
	unless (-x $self->{SCMXX}) {
		my $tmp = qq|ERROR ACCESSING MOBILE PHONE!

Your current Preferences are
    Device:    $self->{DEVICE}
    Baudrate:  $self->{BAUDRATE}
    SCMxx:     $self->{SCMXX}

    Phonebook Timeout:     $self->{PBTIMEOUT} s
    Addressbook Timeout:   $self->{ABTIMEOUT} s

Maybe you should fix your setup.

|;
		return $tmp;		
	}
	my $cmd = "$self->{SCMXX} -d $self->{DEVICE} -b $self->{BAUDRATE} $options ";
	print STDERR "Executing CMD: $cmd\n\n" if $SCMxx::DEBUG;
	my $result;
	my $pid = open (CMD, "$cmd | ") || warn "Can't open $cmd";
	print "scmxx called at PID $pid\n" if $SCMxx::DEBUG > 1;
	eval {
		local $SIG{ALRM} = sub { &__timeout($self, $pid, $self->{TIMEOUT}); };
		alarm($self->{TIMEOUT});
		&{$self->{DOIT}} if ($self->{DOIT});
		while (<CMD>) {
			$result .= $_;
			&{$self->{DOIT}} if ($self->{DOIT});
		}
		alarm(0);
	};
   	if ( $@ ) {
		print "***  \n$cmd timed out\n" if $SCMxx::DEBUG > 1;
		my $tmp =  "Error executing command\n'$cmd'\n\n$@";
		$result = $tmp . $result;
		alarm(0);
		$self->reset; # foreach (0..2); 
	}
	close (CMD);
#	print "RESULT: $result" if $SCMxx::DEBUG;
	return unless $result;
    $result =~ s/\x0D//g;
	$result ||= "";
   	if (
		$result =~ /Error/ || 
		$result =~ /Aborting/ || 
		$result =~ /Unexpected/
		) {
		my $tmp = qq|ERROR ACCESSING MOBILE PHONE!

$result

Your current Preferences are
    Device:    $self->{DEVICE}
    Baudrate:  $self->{BAUDRATE}
    SCMxx:     $self->{SCMXX}

    Phonebook Timeout:     $self->{PBTIMEOUT} s
    Addressbook Timeout:   $self->{ABTIMEOUT} s

Maybe you should fix your setup.

|;
		$result = $tmp;
	} else {
		my @tmp = split (/\n/, $result);
		shift (@tmp) if ($tmp[0] =~ /^Accessing device/);
		shift (@tmp) if ($tmp[0] =~ /^OK, a modem device is present./);
		shift (@tmp) if ($tmp[0] =~ /^Detected./);
		shift (@tmp) if ($tmp[0] =~ /^Receiving:$/);
		$result = join ("\n", @tmp);
	}
	return $result;
}

sub __timeout {
	my $self = shift;
	my $pid = shift;
	my $timeout = shift;
	print "******** timeout called : pid is $pid (gscmxx is $$) **********\n" if $SCMxx::DEBUG > 1;
	kill TERM => $pid;
#	$self->reset foreach (0..2); 
	die "TIMEOUT after $timeout seconds\n";
}

##################################################
# returns the VERSION string of scmxx
##################################################
sub version {
	my $self = shift;
	unless ($self->{VERSION}) {
		$self->{VERSION} = '0.0';
		my $result = $self->docmd("--version");
		if ($result =~ /ERROR/) {
			$self->{VERSION} = 0.0;
		} else {
			$self->{VERSION} = $result;
		}
		chomp $self->{VERSION};
	}
	return $self->{VERSION};
}


### control if phonenumber is valid
sub __control_number {
	my $num = shift;

	$num =~ s/\s+//g;
	if ($num !~ /^[\+?\d\#\*]+$/) {
		return (0,"Sorry wrong format in number '$num' (only '+', '#', '*' and numbers allowed)\n");
	}
	return ($num, 0);
}

# temporary directory
# stolen from CGI.pm
sub __find_tmp_dir {
	my $TMPDIRECTORY;
	my @TEMP=("/usr/tmp","/var/tmp", "/tmp","/temp");
	unshift(@TEMP,$ENV{'TMPDIR'}) if exists $ENV{'TMPDIR'};
	unshift(@TEMP,(eval {(getpwuid($<))[7]}).'/tmp'); #  if $CGI::OS eq 'UNIX';
	foreach (@TEMP) {
		do {$TMPDIRECTORY = $_; last} if -d $_ && -w _;
	}
	return $TMPDIRECTORY
}

sub setdebug {
	my $self = shift;
	my $debug = shift;
	$SCMxx::DEBUG = $debug;
}

1;
__END__
# Below is stub documentation for your module. You better edit it!

=head1 NAME

SCMxx - Perl extension for scmxx

=head1 SYNOPSIS

  use SCMxx;
  my $scmxx = new SCMxx;

  my $scmxxversion = $scmxx->version;
  my $infotxt = $scmxx->information;
  my %available_phonebooks = $scmxx->get_available_phonebooks;
  my @phonebook = $scmxx->get_phonebook($pbook);

  
=head1 DESCRIPTION

Documentation for SCMxx,

=head1 Constructor

=over 4

=item new

This is the constructor for a new SCMxx object. No options are required.

=back

=head1 FUNCTIONS

 

=head2 General Functions

=over 4

=item set_preferences(key => value)

This will set one or more preferences for scmxx. 

=item information

This will return the info text provided by the cellphone as scalar.
(output of scmxx -i)

=item reset

Try to reset the phone (scmxx --reset). Will return STDOUT and STDERR of
scmxx

=item setdebug

Set the debug, will show scmxx commands on STDOUT

=item version

This will return the version of scmxx

=back

=head2 SMS related Functions

=over 4

=item get_sms($slot, $format)

This will return the SMS of a specific slot (or 'all' if not specified).
You can choose the format of the return value. Possible formats are:
 TEXT:  returns the output of scmxx
 ARRAY: returns an array of structured sms-messages 
        (a hash ref for each sms)
 HASH:  returns a hash with slot as key and a the sms 
        as value (as hash ref)

=item delete_sms($slot)

Removes the SMS on Slot $slot from the phone.

=item send_sms($number, $text, [$flash, $saveonly]) 

This will send a SMS to $number with text $text. The optional parameter 
$flash will send a 'flash'-sms (which is displayed directly if the phone
of the receiver supports this option). $saveonly will save the SMS text at
the phone. 

=back

=head2 Phonebook Functions

=over 4

=item get_phonebook($pbook, [$format, $timeout])

This will fetch the specified Phonebook (or SM) from your phone and return
it as ARRAY or HASH containing hash refs {NUMBER => 000, NAME => xxx }

=item create_phonebook_from_data($data, [$format])

This will generate a Phonebook from text. $data must contain multiple lines, each in the following format:
  index,"number","name"\n 
  (eg. 3,"5551234","My Friend"\n)

=item available_phonebooks

Returns a hash. Keys are the short forms of the Phonebook names (eg. SM) and values are the descriptions.

=item generate_phonebook($data)

Generates a Phonebook, which can be sent to scmxx. Format is described above.
It returns an array ($status, $data). If $status is not 0, the generation of 
$data was successfull, else $data is an errorstring.

=item save_phonebook($pbook, $data [,$pin2])

=back

=head2 Addressbook Functions

=over 4

=item get_addressbook([$timeout])

Fetch Addressbook from phone

=item create_addressbook_from_data($data)

Create a HASH from given data

=item create_address($data)

Returns a hash ref with Porperties of vcards as keys and their values as values.


=item delete_address($slot)

Delete vcard from slot $slot. Returns output of scmxx.

=item delete_addressbook_entry($slot)

The same as delete_address.

=item save_address($slot, $address)

Calls generate_vcare($address) and saves the resulting vcard it on the phone at slot $slot.

=item generate_vcard($address)

Generates a vcard from $addess (= a hash ref)

=back

=head2 Bitmap functions

=over 4

=item get_bitmaps($slot [, $bitmapdir, $format])

Fetches bitmap from slot $slot (or all if not specified) and saves it in dir $bitmapdir (or '.'). Filenames are $slot.bmp. If you specify $format = 'xpm' it will convert the bitmap(s) to xpm-format, but leave the original bitmap untouched.
(This requires bmptoppm and ppmtoxpm)
Returns an array with the names of the fetched bitmaps.

=back

=head2 Midi File Functions

=over 4

=item get_sounds($slot [,$dir])

Fetches midi-file from slot $slot (or all if not specified) and saves it in dir $dir (or '.'). Filenames are $slot.mid. 
It returns a array containing the name(s) of the fetched midi-file(s)
(This is very alpha)

=back

=head1 BUGS

A lot, I think.

=head1 TODO

A lot! Eg. save bitmaps and sounds, check input, ....

=head1 AUTHOR

Koppi, koppi@weirdoz.net

=head1 SEE ALSO

scmxx, perl(1).

=cut

